local api = require("gemini.api")
local M = {}

-- State for managing current suggestion and debounce timer
local current_suggestion = {
    text = nil,
    start_col = nil,
    namespace_id = vim.api.nvim_create_namespace('gemini_suggestion'),
    timer = nil,
    last_request = 0,
    cache = {},
    debounce_ms = 1000, -- Wait 1 second between requests
    min_chars = 0,      -- Minimum characters before triggering completion
    max_context_lines = 10, -- Maximum number of context lines to send
}

-- Add at the top of the file with other state variables
local completion_cache = {
    entries = {},
    max_entries = 1000,
    ttl = 30 * 60 * 1000, -- 30 minutes TTL
    last_cleanup = 0
}

-- Add completion triggers detection
local function should_trigger_completion(line_text, col)
    local config = require("gemini.config")
    local trigger_chars = config.options.completion.trigger_characters or "."
    
    -- Don't trigger on empty lines or spaces
    if line_text:match("^%s*$") then return false end
    
    -- Check minimum characters
    local text_before_cursor = line_text:sub(1, col)
    if #text_before_cursor < current_suggestion.min_chars then return false end
    
    -- Check if we're in a comment
    local filetype = vim.bo.filetype
    local comment_string = vim.bo.commentstring
    if comment_string then
        local comment_start = comment_string:match("^(.-)%%s")
        if comment_start and text_before_cursor:match(vim.pesc(comment_start)) then
            return false
        end
    end
    
    -- Check for trigger characters
    local last_char = text_before_cursor:sub(-1)
    if trigger_chars:find(last_char, 1, true) then
        return true
    end
    
    -- Check for meaningful context
    local meaningful_pattern = "[%w_][%w_%.%:]*$" -- Matches identifiers and dot/colon access
    local context = text_before_cursor:match(meaningful_pattern)
    if context and #context >= 2 then
        return true
    end
    
    return false
end

-- Improved cache key generation
local function get_cache_key(context, line_text, col)
    local prefix = line_text:sub(1, col)
    local filetype = vim.bo.filetype
    local key_parts = {
        filetype,
        prefix,
        vim.inspect(context):sub(1, 100) -- Limited context hash
    }
    return vim.fn.sha256(table.concat(key_parts, "||"))
end

-- Cache management
local function cleanup_cache()
    local now = vim.loop.now()
    if now - completion_cache.last_cleanup < 60000 then return end -- Cleanup max once per minute
    
    local count = 0
    local expired = {}
    
    for key, entry in pairs(completion_cache.entries) do
        if now - entry.timestamp > completion_cache.ttl then
            table.insert(expired, key)
        end
        count = count + 1
    end
    
    -- Remove expired entries
    for _, key in ipairs(expired) do
        completion_cache.entries[key] = nil
    end
    
    -- If still too many entries, remove oldest
    if count > completion_cache.max_entries then
        local entries = {}
        for k, v in pairs(completion_cache.entries) do
            table.insert(entries, {key = k, timestamp = v.timestamp})
        end
        table.sort(entries, function(a, b) return a.timestamp < b.timestamp end)
        
        for i = 1, count - completion_cache.max_entries do
            completion_cache.entries[entries[i].key] = nil
        end
    end
    
    completion_cache.last_cleanup = now
end

-- Cache access
local function get_cached_completion(context, line_text, col)
    cleanup_cache()
    local key = get_cache_key(context, line_text, col)
    local entry = completion_cache.entries[key]
    
    if entry and (vim.loop.now() - entry.timestamp) < completion_cache.ttl then
        return entry.completion
    end
    return nil
end

local function cache_completion(context, line_text, col, completion)
    local key = get_cache_key(context, line_text, col)
    completion_cache.entries[key] = {
        completion = completion,
        timestamp = vim.loop.now()
    }
end

-- Helper function to get visible lines around cursor
local function get_visible_lines()
    local win = vim.api.nvim_get_current_win()
    local cursor_line = vim.api.nvim_win_get_cursor(0)[1]
    local top_line = vim.fn.line('w0')
    local bottom_line = vim.fn.line('w$')
    
    -- Get all visible lines
    local lines = vim.api.nvim_buf_get_lines(0, top_line - 1, bottom_line, false)
    local relative_cursor = cursor_line - top_line + 1
    
    return lines, relative_cursor
end

-- Helper function to get current line indent
local function get_line_indent(line)
    local indent = line:match("^%s+") or ""
    return indent
end

-- Debug function
local function debug_print(...)
    vim.notify(string.format(...), vim.log.levels.INFO)
end

-- Helper function to clear suggestion
local function clear_suggestion()
    if current_suggestion.timer then
        vim.fn.timer_stop(current_suggestion.timer)
        current_suggestion.timer = nil
    end
    
    if current_suggestion.text then
        vim.api.nvim_buf_clear_namespace(0, current_suggestion.namespace_id, 0, -1)
        current_suggestion.text = nil
        current_suggestion.start_col = nil
    end
end

-- Check if we should rate limit
local function should_rate_limit()
    local now = vim.loop.now()
    if (now - current_suggestion.last_request) < current_suggestion.debounce_ms then
        return true
    end
    current_suggestion.last_request = now
    return false
end

-- Cache key generation
local function get_cache_key(context)
    return vim.fn.sha256(context)
end

-- Helper function to show suggestion
local function show_suggestion(suggestion, start_col)
    clear_suggestion()
    
    local cursor = vim.api.nvim_win_get_cursor(0)
    local line = cursor[1] - 1
    local current_line = vim.api.nvim_get_current_line()
    
    -- Get current line indent
    local indent = get_line_indent(current_line)
    
    -- Split suggestion into lines and apply indent to all lines except first
    local suggestion_lines = vim.split(suggestion, "\n")
    for i = 2, #suggestion_lines do
        suggestion_lines[i] = indent .. suggestion_lines[i]
    end
    
    if #suggestion_lines == 0 then return end
    
    -- Add support for different suggestion styles
    local config = require("gemini.config")
    local style = config.options.completion_style or "ghost" -- Add this to config

    -- Show ghost text with improved styling
    vim.api.nvim_buf_set_extmark(0, current_suggestion.namespace_id, line, start_col, {
        virt_text = {{suggestion_lines[1], 'GeminiSuggestion'}},
        virt_text_pos = 'overlay',
        hl_mode = 'combine',
        priority = 100, -- Add higher priority to ensure visibility
        right_gravity = false -- Ensure suggestion stays at cursor position
    })
    
    -- Show remaining lines as virtual lines
    if #suggestion_lines > 1 then
        local virt_lines = {}
        for i = 2, #suggestion_lines do
            table.insert(virt_lines, {{suggestion_lines[i], 'GeminiSuggestion'}})
        end
        
        vim.api.nvim_buf_set_extmark(0, current_suggestion.namespace_id, line, start_col, {
            virt_lines = virt_lines,
            virt_lines_above = false,
        })
    end
    
    current_suggestion.text = table.concat(suggestion_lines, "\n")
    current_suggestion.start_col = start_col
end

function M.accept_suggestion()
    if not (current_suggestion.text and current_suggestion.start_col) then
        return false
    end

    local line = vim.api.nvim_win_get_cursor(0)[1] - 1
    local col = vim.api.nvim_win_get_cursor(0)[2]
    
    -- Only accept if we're still at or after the suggestion start
    if col < current_suggestion.start_col then
        return false
    end
    
    -- Get current line text
    local line_text = vim.api.nvim_buf_get_lines(0, line, line + 1, true)[1]
    local text_after_cursor = string.sub(line_text, col + 1)
    
    -- Split suggestion into lines
    local suggestion_lines = vim.split(current_suggestion.text, "\n")
    if #suggestion_lines == 0 then return false end
    
    -- For the first line, only insert the part that doesn't overlap
    local first_line = suggestion_lines[1]
    local common_length = 0
    while common_length < #text_after_cursor and common_length < #first_line do
        if string.sub(text_after_cursor, common_length + 1, common_length + 1) ==
           string.sub(first_line, common_length + 1, common_length + 1) then
            common_length = common_length + 1
        else
            break
        end
    end
    
    -- Only insert the non-overlapping part
    local to_insert = string.sub(first_line, common_length + 1)
    if to_insert ~= "" then
        vim.api.nvim_buf_set_text(
            0,
            line,
            col,
            line,
            col,
            {to_insert}
        )
    end
    
    -- Insert remaining lines below
    if #suggestion_lines > 1 then
        vim.api.nvim_buf_set_lines(
            0,
            line + 1,
            line + 1,
            false,
            vim.list_slice(suggestion_lines, 2)
        )
    end
    
    -- Move cursor to end of inserted text
    local final_line = line + #suggestion_lines - 1
    local final_col
    if final_line == line then
        final_col = col + #to_insert
    else
        final_col = #suggestion_lines[#suggestion_lines]
    end
    vim.api.nvim_win_set_cursor(0, {final_line + 1, final_col})
    
    -- Clear suggestion and cache for the current context
    clear_suggestion()
    current_suggestion.cache = {} -- Clear cache to force new suggestions
    
    -- Schedule next completion after a short delay
    vim.defer_fn(function()
        if vim.b.gemini_completion_enabled then
            M.trigger_completion()
        end
    end, 100)
    
    return true
end

-- Helper function to get relevant context around cursor
local function get_context_around_cursor(lines, current_line, max_lines)
    local context = {}
    local start_line = math.max(1, current_line - math.floor(max_lines/2))
    local end_line = math.min(#lines, current_line + math.floor(max_lines/2))
    
    -- Add file type and cursor position information
    local file_type = vim.bo.filetype
    local cursor_pos = vim.api.nvim_win_get_cursor(0)
    
    -- Add metadata to context
    table.insert(context, string.format("File type: %s", file_type))
    table.insert(context, string.format("Cursor position: line %d, col %d", cursor_pos[1], cursor_pos[2]))
    
    -- Add visible context with line numbers
    for i = start_line, end_line do
        if lines[i] and #lines[i] > 0 then
            table.insert(context, string.format("L%d: %s", i, lines[i]))
        end
    end
    
    return context
end

local function create_debounced_function(fn, wait)
    local timer = nil
    return function(...)
        local args = {...}
        if timer then
            vim.fn.timer_stop(timer)
        end
        timer = vim.fn.timer_start(wait, function()
            fn(unpack(args))
            timer = nil
        end)
    end
end

-- Use the improved debouncing
local trigger_completion_debounced = create_debounced_function(function()
    M.trigger_completion()
end, current_suggestion.debounce_ms)

function M.trigger_completion()
    -- Early exit conditions
    if not vim.b.gemini_completion_enabled then return end
    if vim.fn.pumvisible() ~= 0 then return end
    
    local cursor = vim.api.nvim_win_get_cursor(0)
    local line = cursor[1] - 1
    local col = cursor[2]
    local current_line = vim.api.nvim_get_current_line()
    
    -- Check if we should trigger completion
    if not should_trigger_completion(current_line, col) then
        clear_suggestion()
        return
    end
    
    -- Check rate limiting
    if should_rate_limit() then return end
    
    -- Get context
    local visible_lines = vim.api.nvim_buf_get_lines(0, math.max(0, line - 10), line + 10, false)
    local context = get_context_around_cursor(visible_lines, 10, current_suggestion.max_context_lines)
    
    -- Check cache first
    local cached_completion = get_cached_completion(context, current_line, col)
    if cached_completion then
        vim.schedule(function()
            show_suggestion(cached_completion, col)
        end)
        return
    end
    
    -- Create prompt for completion
    local prompt = string.format([[
You are an autocomplete engine. Respond ONLY with the direct completion text.
DO NOT include:
- Explanations
- Markdown formatting
- Code block markers
- Additional newlines
- Any other text

Language: %s
Context:
%s
Complete this line:
%s]], vim.bo.filetype, table.concat(context, "\n"), current_line)
    
    -- Make API request
    api.get_response(prompt, nil, function(response, error)
        if error then
            if error:match("code = 429") then
                handle_rate_limit()
                return
            end
            return
        end
        
        if response then
            -- Cache the successful completion
            cache_completion(context, current_line, col, response)
            
            vim.schedule(function()
                local new_cursor = vim.api.nvim_win_get_cursor(0)
                if new_cursor[1] - 1 == line and new_cursor[2] == col then
                    show_suggestion(response, col)
                end
            end)
        end
    end)
end

-- Setup function to create highlight group and keymaps
function M.setup()
    -- Create highlight group for suggestions
    vim.api.nvim_set_hl(0, 'GeminiSuggestion', {
        fg = '#666666',
        italic = true,
        blend = 15  -- Makes the ghost text slightly transparent
    })
    
    -- Map Tab to accept suggestion
    vim.keymap.set('i', '<Tab>', function()
        if not M.accept_suggestion() then
            vim.api.nvim_feedkeys(vim.api.nvim_replace_termcodes('<Tab>', true, true, true), 'n', true)
        end
    end, { expr = false, noremap = true })
    
    -- Set up autocommand for real-time completion
    vim.api.nvim_create_autocmd("TextChangedI", {
        pattern = "*",
        callback = function()
            if vim.b.gemini_completion_enabled then
                M.trigger_completion()
            end
        end
    })
    
    -- Clear suggestion on cursor move
    vim.api.nvim_create_autocmd({'CursorMovedI', 'CursorMoved'}, {
        pattern = "*",
        callback = function()
            clear_suggestion()
        end
    })
end

return M
